import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin

try:
    import torch
    import torch.nn as nn
    import torch.optim as optim
    from torch.utils.data import DataLoader, TensorDataset
except ImportError:
    torch = None
    nn = None
    optim = None
    DataLoader = None
    TensorDataset = None


class Autoencoder(nn.Module):
    def __init__(self, encoder_dims, decoder_dims):
        if torch is None or nn is None or optim is None:
            raise ImportError(
                "torch is required for AutoencoderTransformer. Please install it using 'pip install your_package_name[torch]'."
            )
        super(Autoencoder, self).__init__()
        # Create the encoder
        encoder_layers = []
        for i in range(len(encoder_dims) - 1):
            encoder_layers.append(nn.Linear(encoder_dims[i], encoder_dims[i + 1]))
            encoder_layers.append(nn.ReLU())
        self.encoder = nn.Sequential(*encoder_layers)

        # Create the decoder
        decoder_layers = []
        for i in range(len(decoder_dims) - 1):
            decoder_layers.append(nn.Linear(decoder_dims[i], decoder_dims[i + 1]))
            decoder_layers.append(nn.ReLU())
        self.decoder = nn.Sequential(*decoder_layers)

    def forward(self, x):
        x = self.encoder(x)
        x = self.decoder(x)
        return x


class AutoencoderTransformer(BaseEstimator, TransformerMixin):
    def __init__(
        self,
        encoder_dims,
        decoder_dims,
        learning_rate=0.001,
        num_epochs=100,
        batch_size=32,
    ):
        self.input_dim = encoder_dims
        self.hidden_dim = decoder_dims
        self.learning_rate = learning_rate
        self.num_epochs = num_epochs
        self.batch_size = batch_size
        self.autoencoder = Autoencoder(encoder_dims, decoder_dims)

    def fit(self, X, y=None):
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.autoencoder.to(device)

        # Normalize the functional connectivity data
        X_norm = (X - np.mean(X, axis=0)) / np.std(X, axis=0)

        criterion = nn.MSELoss()
        optimizer = optim.Adam(self.autoencoder.parameters(), lr=self.learning_rate)

        data_loader = DataLoader(
            TensorDataset(torch.tensor(X_norm, dtype=torch.float32)),
            batch_size=self.batch_size,
        )

        for epoch in range(self.num_epochs):
            for data in data_loader:
                input_data = data[0].to(device)
                input_data = input_data.reshape(-1, self.input_dim)
                output = self.autoencoder(input_data)

                loss = criterion(output, input_data)

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
        return self

    def transform(self, X, y=None):
        self.autoencoder.eval()
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        X_tensor = torch.tensor(X, dtype=torch.float32).to(device)
        return self.autoencoder.encoder(X_tensor).cpu().detach().numpy()


if __name__ == "__main__":
    fc_data = np.random.randn(
        100, 400
    )  # 400 is the example input dimension, adjust it to match your data
    input_dim = 400
    hidden_dim = 128
    autoencoder_transformer = AutoencoderTransformer(
        input_dim=input_dim, hidden_dim=hidden_dim
    )
    autoencoder_transformer.fit(fc_data)
    fc_data_transformed = autoencoder_transformer.transform(fc_data)



# robust loading functions
import csv
import pandas as pd
import nibabel as nib
import bids


def rbload_csv(csv_file):
    if isinstance(csv_file, str):
        with open(csv_file, "r") as file:
            dialect = csv.Sniffer().sniff(file.read(1024))
            file.seek(0)
            return pd.read_csv(csv_file, delimiter=dialect.delimiter)
    elif isinstance(csv_file, pd.DataFrame):
        return csv_file


def rbload_imgs(
    imgs: list or str or nib.nifti1.Nifti1Image or bids.layout.models.BIDSImageFile,
):
    """
    Load images from a list of paths, a single path, or a list of nibabel images
    Return a list of nibabel images
    """
    if isinstance(imgs, str):
        return [nib.load(imgs)]
    elif isinstance(imgs, nib.nifti1.Nifti1Image):
        return [imgs]
    elif isinstance(imgs, bids.layout.models.BIDSImageFile):
        return imgs.get_image()
    try:
        imgs = list(imgs)
    except TypeError:
        raise ValueError(
            "If imgs is not a single file, then imgs must can be converted to list"
        )
    try:
        if isinstance(imgs[0], str):
            imgs = [nib.load(stat_map) for stat_map in imgs]
        elif isinstance(imgs[0], bids.layout.models.BIDSImageFile):
            imgs = [img.get_image() for img in imgs]
        elif isinstance(imgs[0], nib.nifti1.Nifti1Image):
            pass
        else:
            raise ValueError("imgs must be list of str/nib.Nifti1Image/BIDSImageFile")
    except IndexError:
        print("imgs must be non-empty")

    return imgs



from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score
from sklearn.metrics import (
    accuracy_score,
    precision_score,
    recall_score,
    f1_score,
    roc_auc_score,
    confusion_matrix,
)
from scipy import stats
import numpy as np


def icc(pred, target):
    # Intra-class Correlation Coefficient (ICC)
    n = len(pred)
    mean_pred = np.mean(pred)
    mean_target = np.mean(target)
    ss_total = np.sum(np.square(target - mean_target)) + np.sum(
        np.square(pred - mean_pred)
    )
    ss_error = np.sum(np.square(target - pred))
    icc_val = (ss_total - ss_error) / ss_total
    return icc_val


def evaluate_continuous(pred, target):
    r = stats.pearsonr(pred, target)[0]
    mae = mean_absolute_error(pred, target)
    r2 = r2_score(pred, target)
    mse = mean_squared_error(pred, target)
    rmse = np.sqrt(mse)
    icc_val = icc(pred, target)
    # TODO: Fix this, ,explained_variance_score is not defined
    # explained_variance = explained_variance_score(pred, target)
    metrics = {
        "Pearson r": r,
        "MAE": mae,
        "R2": r2,
        "MSE": mse,
        "RMSE": rmse,
        "ICC": icc_val,
    }

    return metrics


def evaluate_binary(pred, target, threshold=0.5):
    # Convert probability predictions to binary predictions using the threshold
    binary_pred = np.where(np.array(pred) >= threshold, 1, 0)

    accuracy = accuracy_score(target, binary_pred)
    precision = precision_score(target, binary_pred)
    recall = recall_score(target, binary_pred)
    f1 = f1_score(target, binary_pred)
    roc_auc = roc_auc_score(target, pred)  # Use probability predictions for ROC AUC
    cm = confusion_matrix(target, binary_pred)

    metrics = {
        "Accuracy": accuracy,
        "Precision": precision,
        "Recall": recall,
        "F1 Score": f1,
        "ROC AUC": roc_auc,
        "Confusion Matrix": cm,
    }

    return metrics



import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import PolynomialFeatures


class BrainAgeGapCorrectionLinear(BaseEstimator, TransformerMixin):
    #     References:
    #     Bias-adjustment in neuroimaging-based brain age frameworks: A robust scheme
    def __init__(self, line_reg_order=1, out="bag"):
        self.line_reg_order = line_reg_order
        self.line_reg = LinearRegression()
        self.poly = PolynomialFeatures(degree=self.line_reg_order, include_bias=True)
        self.out = out

    def fit(self, X, y):
        line_reg_features = self.poly.fit_transform(y.reshape(-1, 1))
        offset_train = X - y
        self.line_reg.fit(line_reg_features, offset_train.reshape(-1, 1))
        return self

    def transform(self, X, y):
        test_features = self.poly.transform(y.reshape(-1, 1))
        offset_test = np.array(self.line_reg.predict(test_features)).flatten()
        corrected_age = X - offset_test

        if self.out == "age":
            return corrected_age.reshape(-1, 1)
        elif self.out == "bag":
            bag = corrected_age - y
            return bag.reshape(-1, 1)

    def fit_transform(self, X, y):
        return self.fit(X, y).transform(X, y)


def BAG(train_pred, test_pred, train_y, test_y, line_reg_order=1):
    """
    Brain Age Gap Correction

    Parameters:
    train_pred (1D numpy array): Predicted age on the training set.
    test_pred (1D numpy array): Predicted age on the test set.
    train_y (1D numpy array): Actual age in the training set.
    test_y (1D numpy array): Actual age in the test set.

    Returns:
    corrected_age (numpy array): Corrected predicted age on the test set.
    train_delta_age (numpy array): Difference between predicted and actual age on the training set.
    line_reg (LinearRegression object): Trained regression model.

    """

    line_reg = LinearRegression()
    poly = PolynomialFeatures(degree=line_reg_order, include_bias=True)
    line_reg_features = poly.fit_transform(train_y.reshape(-1, 1))
    offset_train = train_pred - train_y
    line_reg.fit(line_reg_features, offset_train.reshape(-1, 1))
    test_features = poly.transform(test_y.reshape(-1, 1))
    offset_test = np.array(line_reg.predict(test_features)).flatten()
    corrected_age = test_pred - offset_test
    corrected_delta_age = corrected_age - test_y
    return corrected_age, corrected_delta_age



import os
import tqdm
import warnings
import pandas as pd
import nibabel as nib
import warnings

# def _copy_image(image):
#     image_data_copy = image.get_fdata().copy()
#     affine_copy = image.affine.copy()
#     return nib.Nifti1Image(image_data_copy, affine_copy)


def install_workbench(path, version="1.5.0", redhat: bool = "warn"):
    import platform
    import zipfile
    import urllib.request
    from pathlib import Path

    workbench_urls = {
        "Linux": f"https://humanconnectome.org/storage/app/media/workbench/workbench-linux64-v{version}.zip",
        "RedHat": f"https://humanconnectome.org/storage/app/media/workbench/workbench-rh_linux64-v{version}.zip",
        "Darwin": f"https://humanconnectome.org/storage/app/media/workbench/workbench-mac64-v{version}.zip",
        "Windows": f"https://humanconnectome.org/storage/app/media/workbench/workbench-windows64-v{version}.zip",
    }

    os_type = platform.system()
    if os_type == "Linux":
        distribution = platform.freedesktop_os_release()["ID"]
        if distribution in ["rhel", "centos", "fedora", "red hat"]:
            os_type = "RedHat"
        elif distribution in ["ubuntu", " debian", "arch", "opensuse", "sles"]:
            os_type = "Linux"
        else:
            if redhat == "warn":
                warnings.warn(
                    "Your distribution is not support by this function, but you can choose whether this is a RedHat machine or not"
                )
                return
            elif redhat is True:
                os_type = "RedHat"
            elif redhat is False:
                os_type = "Linux"

    workbench_url = workbench_urls.get(os_type, None)

    if workbench_url is None:
        raise ValueError(f"Unsupported operating system: {os_type}")

    workbench_zip = f"workbench-{os_type.lower()}-v{version}.zip"
    workbench_dir = Path(path)

    print(f"Downloading {workbench_zip}...")
    urllib.request.urlretrieve(workbench_url, workbench_zip)

    print(f"Extracting {workbench_zip} to {workbench_dir}...")
    with zipfile.ZipFile(workbench_zip, "r") as zip_ref:
        zip_ref.extractall(workbench_dir)

    bin_paths = {
        "Linux": "bin_linux64",
        "RedHat": "bin_rh_linux64",
        "Darwin": "bin_darwin64",
        "Windows": "bin_windows64",
    }
    os.environ["PATH"] += f":{path}/workbench/" + bin_paths[os_type]


def compare_all_neuromaps(
    user_map: str,
    src_space="MNI152",
    MNI_trg_space=("fsaverage", "10k"),
    skip_source=None,
    return_nulls=True,
    method="linear",
):
    """
    This function compares a user map to all available neuromaps returned by `neuromaps.datasets.available_annotations()`
    If the neuromap is in surface

    """
    # TODO: change to make use of additional requirements
    try:
        import neuromaps
    except ImportError:
        raise ImportError("neuromaps package is not found. Try `pip install neuromaps`")
    from neuromaps import transforms, stats, images, nulls
    from neuromaps.datasets import available_annotations, fetch_annotation
    from neuromaps.resampling import resample_images

    if skip_source is None:
        skip_source = []
    result = []
    for annotation in tqdm.tqdm(available_annotations()):
        state_map = user_map  # use rbloads()
        source, map_name, space, resolution = annotation
        annotation = list(annotation)
        if source in skip_source:
            continue
        neuro_map = fetch_annotation(source=source, desc=map_name)
        if space == "MNI152":
            space, resolution = MNI_trg_space
            state_map, neuro_map = resample_images(
                state_map,
                neuro_map,
                src_space=src_space,
                trg_space=space,
                method=method,
                resampling="transform_to_alt",
                alt_spec=(space, resolution),
            )

        elif space == "fsaverage":
            state_map = transforms.mni152_to_fsaverage(state_map, resolution)
        elif space == "fsLR":
            if resolution in ["4k"]:
                warnings.warn("resampling to fsLR 4k is not supported, setting to 32k")
                resolution = "32k"
                state_map, neuro_map = resample_images(
                    state_map,
                    neuro_map,
                    src_space=src_space,
                    trg_space=space,
                    method=method,
                    resampling="transform_to_alt",
                    alt_spec=(space, resolution),
                )
            else:
                state_map = transforms.mni152_to_fslr(state_map, resolution)

        elif space == "civet":
            state_map = transforms.mni152_to_civet(state_map, civet_density=resolution)
        else:
            raise ValueError(
                f"space {space} not implemented",
            )

        rotated = nulls.alexander_bloch(
            neuro_map, atlas=space, density=resolution, n_perm=100, seed=1234
        )
        if return_nulls:
            corr, pval, nulls = stats.compare_images(
                state_map, neuro_map, nulls=rotated, return_nulls=return_nulls
            )
            annotation.extend([corr, pval, nulls])
            cols = ["source", "map", "space", "resolution", "corr", "pval", "null"]
        else:
            corr, pval = stats.compare_images(
                state_map, neuro_map, nulls=rotated, return_nulls=return_nulls
            )
            annotation.extend([corr, pval])
            cols = ["source", "map", "space", "resolution", "corr", "pval"]
        result.append(annotation)
    result_df = pd.DataFrame(result)
    result_df.columns = cols
    return result_df


def pretrained_neurosynth_model(method="correlation", save_path="./"):
    import gzip
    import pickle
    import requests
    import os

    # Download and load the decoder
    url = f"https://raw.githubusercontent.com/wetiqe/neuroimager/main/asset/decoder/neurosynth/{method}_decoder.pkl.gz"
    response = requests.get(url)

    if response.status_code == 200:
        if save_path:
            decoder_path = os.path.join(save_path, f"{method}_decoder.pkl.gz")
            with open(decoder_path, "wb") as f:
                f.write(response.content)
            with gzip.open(decoder_path, "rb") as f:
                decoder = pickle.load(f)
        else:
            with gzip.open(response.content, "rb") as f:
                decoder = pickle.load(f)
    else:
        print(
            f"Error: Unable to download the file (status code {response.status_code})"
        )

    # Download and load the features
    url = f"https://raw.githubusercontent.com/wetiqe/neuroimager/main/asset/decoder/neurosynth/selected_features.txt"
    response = requests.get(url)

    if response.status_code == 200:
        if save_path:
            features_path = os.path.join(save_path, "selected_features.txt")
            with open(features_path, "wb") as f:
                f.write(response.content)
            with open(features_path, "r") as f:
                features = f.read().splitlines()
        else:
            features = response.text.splitlines()
    else:
        print(
            f"Error: Unable to download the file (status code {response.status_code})"
        )

    return decoder, features



import numpy as np
import math
from sklearn.base import BaseEstimator, TransformerMixin


# Part one: sklearn estimator types
# These are the wrappers for the functions in part two
class FCFilter(BaseEstimator, TransformerMixin):
    def __init__(self, network_labels, excludes=None, includes=None):
        self.network_labels = network_labels
        self.excludes = excludes
        self.includes = includes

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        return filter_fcs(X, self.network_labels, self.excludes, self.includes)


class BipartiteMatrixExtractor(BaseEstimator, TransformerMixin):
    def __init__(self, network_labels, bp1_labels, bp2_labels):
        self.network_labels = network_labels
        self.bp1_labels = bp1_labels
        self.bp2_labels = bp2_labels

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        return extract_bipartite_matrix(
            X, self.network_labels, self.bp1_labels, self.bp2_labels
        )


class AverageNodesTransformer(BaseEstimator, TransformerMixin):
    def __init__(self, labels):
        self.labels = labels

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        return average_nodes(X, self.labels)


class LowerTriangular(BaseEstimator, TransformerMixin):
    def __init__(self, diagonal=False):
        self.diagonal = diagonal

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        return flatten_lower_triangular(X, self.diagonal)

    def inverse_transform(self, X, y=None):
        return unflatten_lower_triangular(X, self.diagonal)


# Part two: functions
def filter_fcs(
    fcs: np.array, network_labels: list, excludes: list = None, includes: list = None
):
    """Filter FC matrix based on network labels
    :param fcs: A numpy matrix shaped (n_subjects, n_nodes, n_nodes) representing the functional connectivity data.
    :param network_labels: A list containing the network labels for each node.
    :param excludes: A list of network labels to exclude from the filtered FC matrix. Default is None.
    :param includes: A list of network labels to include in the filtered FC matrix. Default is None.
    :return: A filtered numpy matrix containing only the desired network labels.

    This function filters the input functional connectivity matrix based on the specified network labels.
    It can be used to include or exclude specific networks from the analysis.
    """
    if includes is not None and excludes is not None:
        raise ValueError("Only one of 'includes' and 'excludes' can be specified.")
    elif excludes is None:
        excludes = []
    elif includes is None:
        includes = []

    def get_index(labels: list, desired: list):
        return [i for i, label in enumerate(labels) if label in desired]

    if includes:
        ind = get_index(network_labels, includes)
        indexer = np.ix_(np.arange(fcs.shape[0]), ind, ind)
        fcs = fcs[indexer]
    elif excludes:
        ind = get_index(network_labels, excludes)
        fcs = np.delete(fcs, ind, axis=1)
        fcs = np.delete(fcs, ind, axis=2)

    return fcs


def extract_bipartite_matrix(
    matrix: np.array, network_labels: list, bp1_labels: list, bp2_labels: list
):
    """Extract the bipartite part of a functional connectivity matrix.

    :param matrix: A numpy matrix shaped (n_subjs, n_nodes, n_nodes) representing the functional connectivity data.
    :param network_labels: A list containing the network labels for each node.
            should be the same order and length as the number of nodes in the matrix.
    :param bp1_labels: A list of network names for Network 1 nodes.
    :param bp2_labels: A list of network names for other networks' nodes.
    :return: A numpy matrix containing the desired bipartite part of the functional connectivity matrix.
    """
    # the network_labels should have same length as nodes in the matrix
    if len(network_labels) != matrix.shape[1]:
        raise ValueError(
            "The length of network_labels should be equal to the number of nodes in the matrix"
        )

    network1_indices = np.isin(network_labels, bp1_labels)
    other_networks_indices = np.isin(network_labels, bp2_labels)

    bipartite_matrix = matrix[
        np.ix_(range(matrix.shape[0]), network1_indices, other_networks_indices)
    ]

    return bipartite_matrix


def average_nodes(fc_matrix, labels):
    """
    Computes an averaged functional connectivity matrix for each unique label
    in the input labels array. The function accepts both 2D and 3D fc_matrix arrays.

    Parameters:
    -----------
    fc_matrix : numpy array, shape (subject, node, node) or (node, node)
        Functional connectivity matrix or matrices. If 2D, it is assumed to be a single
        matrix. If 3D, each matrix along the first axis corresponds to a subject.

    labels : numpy array, shape (subject,)
        Labels corresponding to each subject in the fc_matrix. Each unique label
        represents a group for which an averaged functional connectivity matrix
        will be computed.

    Returns:
    --------
    avg_fc_matrices : numpy array, shape (unique_labels, node, node)
        Averaged functional connectivity matrices for each unique label in the
        input labels array. If the input fc_matrix is 2D, the output will also
        be 2D with shape (node, node).
    """
    if len(fc_matrix.shape) == 2:
        fc_matrix = np.expand_dims(fc_matrix, axis=0)
    unique_labels = np.unique(np.array(labels))
    n_subjs, n_nodes, _ = fc_matrix.shape
    averaged_fc_matrix = np.zeros((n_subjs, len(unique_labels), len(unique_labels)))

    for i, label in enumerate(unique_labels):
        label_indices = np.where(labels == label)[0]
        for j, other_label in enumerate(unique_labels):
            other_label_indices = np.where(labels == other_label)[0]
            averaged_fc_matrix[:, i, j] = fc_matrix[:, label_indices, :][
                :, :, other_label_indices
            ].mean(axis=(1, 2))

    return averaged_fc_matrix.squeeze()


def flatten_lower_triangular(matrix: np.ndarray, diagonal: bool = False):
    """
    This function takes a 2D or 3D numpy array returns a 2D numpy array.
    where each row corresponds to the flattened lower triangular part (diagonal removed) of a matrix in the input.

    Parameters:
    matrix (numpy.ndarray): A 2D or 3D numpy array representing one or multiple square matrices.

    Returns:
    numpy.ndarray:  A 2D numpy array representing one or multiple flattened lower triangular matrices (diagonal removed)
    """
    if len(matrix.shape) == 2:
        matrix = np.array([matrix])
    elif len(matrix.shape) != 3:
        raise ValueError("Input matrix must be 2D or 3D")
    k = 0 if diagonal else -1
    tri_matrices = np.array([np.tril(fc, k=k) for fc in matrix]).reshape(
        matrix.shape[0], -1
    )
    flat_matrices = tri_matrices[:, ~np.all(tri_matrices == 0, axis=0)]

    return flat_matrices


def unflatten_lower_triangular(flat_matrices, diagonal: bool = False):
    """
    Reconstructs the lower triangular matrices from their flattened representation.

    Args:
        flat_matrices: A 1D or 2D numpy array representing flattened lower triangular matrices (without diagonal).
    Returns:
        A 3D numpy array representing one or multiple lower triangular matrices in their original shape.

    Raises:
        ValueError: If the input matrix is not 2D or has an invalid shape,
                    or if the quadratic equation for `n` has no real solution.
    """
    flat_matrices = np.atleast_2d(flat_matrices)
    if len(flat_matrices.shape) != 2:
        raise ValueError("Input matrix must be 2D")
    m, n = flat_matrices.shape
    k = 0 if diagonal else -1

    def calculate_orig_n(n_nodes):
        discriminant = 1 + 8 * n_nodes
        if discriminant < 0:
            raise ValueError("No real solution for n")
        else:
            n1 = (1 + math.sqrt(discriminant)) / 2
            n2 = (1 - math.sqrt(discriminant)) / 2
            return n1, n2

    n = int(calculate_orig_n(n)[0])
    tri_indices = np.tril_indices(n, k=k)
    original_shape_matrices = np.zeros((m, n, n))

    for i in range(m):
        original_shape_matrices[i][tri_indices] = flat_matrices[i, :]

    if m == 1:
        return original_shape_matrices[0]
    else:
        return original_shape_matrices



from neuroimager.plotting.plot import (
    density_scatter,
    plot_time_series,
    get_sig,
    plot_time_series,
)

from neuroimager.plotting.styler import (
    large_font_size,
    small_font_size,
    no_edge,
)

__all__ = [
    "density_scatter",
    "plot_time_series",
    "get_sig",
    "plot_time_series",
    "large_font_size",
    "small_font_size",
    "no_edge",
]



import numpy as np
from matplotlib import pyplot as plt
import seaborn as sns
from scipy.stats import gaussian_kde


def get_sig(sig):
    symbol = ""
    if 0.01 < sig < 0.05:
        symbol = "*"
    elif 0.001 < sig <= 0.01:
        symbol = "**"
    elif sig <= 0.001:
        symbol = "***"
    elif 0.05 <= sig < 0.10:
        symbol = "^"

    return sig, symbol


def density_scatter(
    x, y, reg=True, gaussian_density=True, figsize=(10, 10), display=True, **kwargs
):
    """
    This function creates a scatter plot with density-based coloring for two input arrays (x and y).
    The density can be calculated using Gaussian kernel density estimation or by converting a 2D histogram to densities.
    Optionally, a linear regression line can be added to the plot.

    Parameters:
    x (array-like): The input data for the x-axis.
    y (array-like): The input data for the y-axis.
    reg (bool, optional): If True, adds a linear regression line to the scatter plot. Default is True.
    gaussian_density (bool, optional): If True, uses Gaussian kernel density estimation to calculate the density.
        If False, calculates the density from a 2D histogram.
        Default is True.
    figsize (tuple, optional): A tuple specifying the width and height of the figure in inches. Default is (10, 10).
    display (bool, optional): If True, displays the plot. Default is True.
    **kwargs: Additional keyword arguments to be passed to the matplotlib scatter plot function.

    Returns:
    fig (matplotlib.figure.Figure): The created Figure object.
    ax (matplotlib.axes.Axes): The created Axes object containing the scatter plot.

    Examples:
    x = np.random.randn(100)
    y = 2 * x + np.random.randn(100)
    fig, ax = density_scatter(x, y, reg=True, gaussian_density=True, figsize=(8, 8))

    Note:
    This function requires the following external libraries: NumPy, Matplotlib, and Seaborn.
    """
    xy = np.vstack([x, y])
    if gaussian_density:
        try:
            from numpy.linalg import LinAlgError

            z = gaussian_kde(xy)(xy)
        except LinAlgError:
            import warnings

            warnings.warn(
                "raised LinAlgError: The covariance matrix associated with the data is singular. "
                "The density plot will be computed from the histogram"
            )
            gaussian_density = False

    if not gaussian_density:
        H, x_edges, y_edges = np.histogram2d(x, y)
        x_centers = (x_edges[:-1] + x_edges[1:]) / 2
        y_centers = (y_edges[:-1] + y_edges[1:]) / 2

        # Convert the histogram to a 2D array of densities
        dx = x_centers[1] - x_centers[0]
        dy = y_centers[1] - y_centers[0]
        z = H / (dx * dy)
    # Create the scatter plot with density-based coloring
    fig, ax = plt.subplots(figsize=figsize)
    sc = ax.scatter(x=x, y=y, c=z, cmap="viridis", **kwargs)
    if reg:
        sns.regplot(x=x, y=y, scatter=False, ax=ax, color="black")

    # Add a colorbar to the figure
    cbar = plt.colorbar(sc)
    cbar.set_label("Density")
    if display:
        plt.show()

    return fig, ax


def plot_time_series(
    time_series,
    n_samples=None,
    y_tick_values=None,
    plot_kwargs=None,
    fig_kwargs=None,
    ax=None,
    filename=None,
):
    """Plot a time series with channel separation.
    ***This is a function modified from else where. Add the source***
    Parameters
    ----------
    time_series : numpy.ndarray
        The time series to be plotted. Shape must be (n_samples, n_channels).
    n_samples : int
        The number of time points to be plotted.
    y_tick_values:
        Labels for the channels to be placed on the y-axis.
    fig_kwargs : dict
        Keyword arguments to be passed on to matplotlib.pyplot.subplots.
    plot_kwargs : dict
        Keyword arguments to be passed on to matplotlib.pyplot.plot.
    ax : matplotlib.axes.Axes
        The axis on which to plot the data. If not given, a new axis is created.
    filename : str
        Output filename.
    Returns
    -------
    fig : matplotlib.pyplot.figure
        Matplotlib figure object.
    ax : matplotlib.pyplot.axis.
        Matplotlib axis object(s).
    """
    time_series = np.asarray(time_series)
    n_samples = min(n_samples or np.inf, time_series.shape[0])
    n_channels = time_series.shape[1]

    # Validation
    if ax is not None:
        if filename is not None:
            raise ValueError(
                "Please use plotting.save() to save the figure instead of the "
                + "filename argument."
            )
        if isinstance(ax, np.ndarray):
            raise ValueError("Only pass one axis.")

    default_fig_kwargs = {"figsize": (12, 8)}
    if fig_kwargs is None:
        fig_kwargs = default_fig_kwargs
    else:
        # fig_kwargs = override_dict_defaults(default_fig_kwargs, fig_kwargs)
        pass
    default_plot_kwargs = {"lw": 0.7, "color": "tab:blue"}
    if plot_kwargs is None:
        plot_kwargs = default_plot_kwargs
    else:
        # plot_kwargs = override_dict_defaults(default_plot_kwargs, plot_kwargs)
        pass
    # Calculate separation
    separation = (
        np.maximum(time_series[:n_samples].max(), time_series[:n_samples].min()) * 1.2
    )
    gaps = np.arange(n_channels)[::-1] * separation

    # Create figure
    if ax is None:
        fig, ax = plt.figure(**fig_kwargs)

    # Plot data
    ax.plot(time_series[:n_samples] + gaps[None, :], **plot_kwargs)

    ax.autoscale(tight=True)
    for spine in ax.spines.values():
        spine.set_visible(False)

    # Set x and y axis tick labels
    ax.set_xticks([])
    if y_tick_values is not None:
        ax.set_yticks(gaps)
        ax.set_yticklabels(y_tick_values)
    else:
        ax.set_yticks([])

    # Save figure
    # if filename is not None:
    #     save(fig, filename)

    return fig, ax



import numpy as np
import pingouin as pg
import pandas as pd


def partial_corrs(
    data,
    x_labels: list = None,
    y_labels: list = None,
    covars: list = None,
    method="auto",
    semi=False,
    alternative="two-sided",
    correct_p="global",
    correct_method="holm",
):
    """
    Compute partial correlations between pairs of variables, controlling for specified covariates.

    Parameters
    ----------
    data : pd.DataFrame or np.ndarray
        A 2D Pandas DataFrame or NumPy Array containing the data.
    x_labels : list
        A list of column indices or names for the independent variables.
    y_labels : list
        A list of column indices or names for the dependent variables.
    covariates : list, optional
        A list of column indices or names for the covariates to control for. Default is None.
    method : str, optional
        The correlation method to use. Options are 'auto', 'pearson', or 'spearman'. Default is 'auto'.
    semi : bool or str, optional
        If True, compute semi-partial correlation.
        If 'x', control for covariates only in x.
        If 'y', control for covariates only in y. Default is False.
    alternative : str, optional
        The alternative hypothesis for the correlation test. Options are 'two-sided', 'greater', or 'less'.
        Default is 'two-sided'.
    correct_p : str, optional
        The correction method for multiple comparisons. Options are 'global', 'IVs', or any other value for no correction.
        Default is 'global'.
    correct_method : str, optional
        The method to use for multiple comparison correction.
        Options are 'holm', 'bonferroni', 'fdr_bh', 'fdr_by', 'fdr_tsbh', 'fdr_tsbky', 'sidak', or 'none'.
        Default is 'holm'.

    Returns
    -------
    r_matrix : pd.DataFrame
        A DataFrame containing the partial correlation coefficients between the independent and dependent variables.
    p_matrix : pd.DataFrame
        A DataFrame containing the p-values for the partial correlation coefficients.

    Notes
    -----
    This function computes partial correlations between pairs of variables, controlling for specified covariates.
    P-values are corrected for multiple comparisons using the specified method.
    The function supports both Pearson and Spearman correlation methods, as well as semi-partial correlations.

    References
    ----------
    1. Pingouin documentation: https://pingouin-stats.org/generated/pingouin.partial_corr.html
    2. Pingouin multiple comparison correction: https://pingouin-stats.org/generated/pingouin.multicomp.html
    """
    if not isinstance(data, pd.DataFrame):
        try:
            data = pd.DataFrame(data)
        except:
            raise ValueError("data must be a 2D Pandas DataFrame or NumPy Array")
    if len(x_labels) == 0 or len(y_labels) == 0:
        raise ValueError("X and Y must be specified")
    if len(covars) == 0:
        print(
            "You did not specify any covariates, will proceed with simple corrlelation"
        )
    IVs = data.columns[x_labels]  # independent variables
    DVs = data.columns[y_labels]  # dependent variables
    r_matrix = pd.DataFrame(np.zeros((len(IVs), len(DVs))), index=IVs, columns=DVs)
    p_matrix = r_matrix.copy()

    for y_label in DVs:
        p_vals = []
        for x_label in IVs:
            # Create a df only contains the variables using now
            # This is a good practice when using pingouin
            df = data[[x_label, y_label] + covars]
            df = df.dropna(how="any", axis=0)  # drop any row with missing value

            # Select correlation method based on multivariate normality test result
            if method == "auto":
                method_to_use = (
                    "pearson"
                    if pg.multivariate_normality(df[[x_label, y_label]])[2]
                    else "spearman"
                )
            else:
                method_to_use = method
            if not semi:
                results = pg.partial_corr(
                    data=df,
                    x=x_label,
                    y=y_label,
                    covar=covars,
                    alternative=alternative,
                    method=method_to_use,
                ).round(3)
            elif semi == "x":
                results = pg.partial_corr(
                    data=df,
                    x=x_label,
                    y=y_label,
                    x_covar=covars,
                    alternative=alternative,
                    method=method_to_use,
                ).round(3)
            elif semi == "y":
                results = pg.partial_corr(
                    data=df,
                    x=x_label,
                    y=y_label,
                    y_covar=covars,
                    alternative=alternative,
                    method=method_to_use,
                ).round(3)

            r_matrix.loc[x_label, y_label] = results.loc[method_to_use, "r"]
            p_vals.append(results.loc[method_to_use, "p-val"])
        if correct_p == "IVs":
            bools, p_vals = pg.multicomp(p_vals, alpha=0.05, method=correct_method)
        p_matrix.loc[:, y_label] = p_vals
    if correct_p == "global":
        pvals = p_matrix.values.flatten()
        bools, p_vals = pg.multicomp(p_vals, alpha=0.05, method=correct_method)
        p_matrix = pd.DataFrame(
            p_vals.reshape(len(x_labels), len(y_labels)),
            index=p_matrix.index,
            columns=p_matrix.columns,
        )
    return r_matrix, p_matrix



import pandas as pd


# explore a give dataframe. These are scripts from an author on kaggle which I forgot the name.
# TODO: I will add the name once I find it.


def Common_data_analysis(
    data,
    missing_value_highlight_threshold=10.0,
    display_df=True,
    only_show_missing=False,
):
    """
    Analsis the common data type and give a table of info such as MMM (mean median mode) ,sd,var etc.

    :param data: dataframe.
    :param missing_value_highlight_threshold: Float the column having missing value percentage above this will be highlighted.

    return: list of all columns and list of categorical column and list of numerical column.
    """
    print("{:=^100}".format(" Common data analysis "))
    # default settings.
    column = data.columns
    total_samples = data.shape[0]
    value_dict = {}

    # calculate values.
    missing_values = data.isnull().sum().values
    missing_value_percentage = [
        round((col_missing_count / total_samples) * 100, 2)
        for col_missing_count in missing_values
    ]
    datatype = [data.iloc[:, i].dtype for i in range(data.shape[1])]

    categorical_data = list(data.loc[:, data.dtypes == "object"].columns)
    numerical_data = [d for d in column if d not in categorical_data]
    # print the diff datatype and count.
    print()
    print(
        "Numerical data list {} ---> total {} numerical values".format(
            numerical_data, len(numerical_data)
        )
    )
    print(
        "Categorical data list {} ---> total {} categorical values".format(
            categorical_data, len(categorical_data)
        )
    )
    print()

    # find is there any special character that represents the missing value.
    missing_value_by_character = [
        data.loc[data[col].isin(["?", " ", "/", "[", "]", "na", "null"]), col]
        .value_counts()
        .sum()
        for col in column
    ]
    missing_value_percentage_by_char = [
        round(col_missing_count / total_samples * 100, 2)
        for col_missing_count in missing_value_by_character
    ]
    # organise values.
    value_dict["data type"] = datatype
    value_dict["Missing Value(NA)"] = missing_values
    value_dict["?[]na null ' ' "] = missing_value_by_character
    value_dict["% of Missing value(NA)"] = missing_value_percentage
    value_dict["% of Missing value(?[]na null ' ')"] = missing_value_percentage_by_char
    df = pd.DataFrame(value_dict, columns=value_dict.keys(), index=column)

    # make a highlight for col has high missing value percentage. (>55% say)
    # the particular row will be highlighted if it above missing value threshold.
    def highlight_high_missing_value(sample):
        threshold = missing_value_highlight_threshold
        style = sample.copy()
        highlight = "background-color: red;"
        if sample[3] > threshold or sample[4] > threshold:
            style[:] = highlight
        else:
            style[:] = ""
        return style

    df_styled = df.style.apply(highlight_high_missing_value, axis=1)
    if display_df:
        if only_show_missing:
            df = df[df["Missing Value(NA)"] > 0]
            display(df)
        else:
            display(df_styled)
    return (column, categorical_data, numerical_data, df)


# columns, categorical_col, numerical_col  = Common_data_analysis(data, 5.0)
#
def numerical_data_analysis(data, numerical_data):
    """
    It will provide the analysis of the numerical data.

    :param data: dataframe.
    :param numerical_data: list of numerical column names.
    """
    print("{:=^100}".format(" Numerical data analysis "))
    print("The skewness we are calculatinf here has a value range ===")
    print("If skew is == 0 -----------> Perfect normal distribution(green color)")
    print("If skew is > +1 -----------> Highly positive skew(red color)")
    print("If skew is > -1 -----------> Highly negative skew(red color)")
    print("If  0 > skew < 0.5 --------> Moderate positive skew(blue color)")
    print("If  0 < skew > -0.5 -------> Moderate negative skew(blue color)")
    column = data.columns

    min_value = [data[col].min() if col in numerical_data else "NA" for col in column]
    max_value = [data[col].max() if col in numerical_data else "NA" for col in column]
    # mode_value = [data[col].mode() if col in numerical_data else "NA" for col in column]
    mean_value = [data[col].mean() if col in numerical_data else "NA" for col in column]
    std_value = [data[col].std() if col in numerical_data else "NA" for col in column]
    # print(mode_value)
    skewness_value = [
        data[col].skew() if col in numerical_data else "NA" for col in column
    ]
    kurtosis_value = [
        data[col].kurtosis() if col in numerical_data else "NA" for col in column
    ]

    q1_value = [
        data[col].quantile(0.25) if col in numerical_data else "NA" for col in column
    ]
    q2_meadian_value = [
        data[col].quantile(0.50) if col in numerical_data else "NA" for col in column
    ]
    q3_value = [
        data[col].quantile(0.75) if col in numerical_data else "NA" for col in column
    ]

    # find the range value.
    def find_range(min_value_list, max_value_list):
        range_value = [
            (max_value - min_value) if min_value != "NA" else "NA"
            for max_value, min_value in zip(max_value_list, min_value_list)
        ]
        return range_value

    # find the inter quartile range. (q3-q1)
    def iqr(q1_value_list, q3_value_list):
        range_value = [
            (q3 - q1) if q1 != "NA" else "NA"
            for q3, q1 in zip(q3_value_list, q1_value_list)
        ]
        return range_value

    range_value = find_range(min_value, max_value)
    iqr_value = iqr(q1_value, q3_value)

    # organise everything inside a dataframe.
    df_dict = {}
    df_dict["min"] = min_value
    df_dict["max"] = max_value
    df_dict["range(max-min)"] = range_value
    # df_dict["mode"] = mode_value
    df_dict["mean/average"] = mean_value
    df_dict["standard deviation"] = std_value
    df_dict["Q1"] = q1_value
    df_dict["meadian/Q2"] = q2_meadian_value
    df_dict["Q3"] = q3_value
    df_dict["Inter quantile range"] = iqr_value
    df_dict["kurtosis"] = kurtosis_value
    df_dict["Skewness"] = skewness_value

    df = pd.DataFrame(df_dict, columns=df_dict.keys(), index=column)

    # highlight the data based on its skewness.
    def highlight_skewness(sample):
        # make a style as the sample shape and property.
        style = sample.copy()
        # make other cell_value as empty style , because i am focusing in coloring skewness column only.
        style[:] = ""
        # set the colors for skewness cells.
        highly_skewed = "background-color: red;"
        moderatly_skewed = "background-color: blue;"
        perfect_normal_destribution = "background-color: green;"

        # color the cells
        if sample[-1] > 1 or sample[-1] < -1:
            style[-1] = highly_skewed
        elif (sample[-1] > 0.5 or sample[-1] <= 1) or (
            sample[-1] > -0.5 or sample[-1] <= -1
        ):
            style[-1] = moderatly_skewed
        elif sample[-1] == 0:
            style[:] = perfect_normal_destribution
        else:
            style[:] = ""
        return style

    df = df.style.apply(highlight_skewness, axis=1)
    display(df)


# Find how much outliers it has using the IQR methods(if data skewed) and z-score method(if data is normally distributed).
def find_outlier_z_score_method(
    data, new_feature=False, col_name=None, return_limits=False
):
    """
    Find the outliers in the given dataset.
    :param data: dataset has number of features to find the outliers.
    :param new_feature: If True create a new feature in the dataFrame to indicate this sample has any outlier feature else do nothing.
    :param return_limit: If true return the upper and lower limits.

    :return data with new feature and number of outlier in each features if new_feature is True, else only number of outlier in each features.
    """
    df = data.copy()
    mean_each_features = df.mean(axis=0)
    std_each_features = df.std(axis=0)
    lower_limit_each_feature = mean_each_features - (3 * std_each_features)
    upper_limit_each_feature = mean_each_features + (3 * std_each_features)

    # find the data is a outlier value or not.
    # print(lower_limit_each_feature, upper_limit_each_feature)
    outlier_df = (df > upper_limit_each_feature) | (df < lower_limit_each_feature)
    # find the number of outliers per feature.
    number_of_outlier_each_feature = outlier_df.sum(axis=0)
    if df.ndim == 1:
        # if the given data is features than handle diffrently. because it has no .values and .index function.
        number_of_outlier_each_feature_df = pd.DataFrame(
            {
                "Feature": [col_name if col_name else "Given feature"],
                "Number of outliers": number_of_outlier_each_feature,
            }
        )
    else:
        number_of_outlier_each_feature_df = pd.DataFrame(
            {
                "Features": number_of_outlier_each_feature.index,
                "Number of outliers": number_of_outlier_each_feature.values,
            }
        )

    # find the upper limit and lower limit Dataframe.
    if df.ndim == 1:
        # if the given data is features than handle diffrently. because it has no .values and .index function.
        lower_limit_each_feature_df = pd.DataFrame(
            {
                "Feature": [col_name if col_name else "Given feature"],
                "lower limit": lower_limit_each_feature,
            }
        )
        upper_limit_each_feature_df = pd.DataFrame(
            {
                "Feature": [col_name if col_name else "Given feature"],
                "upper limit": upper_limit_each_feature,
            }
        )
    else:
        lower_limit_each_feature_df = pd.DataFrame(
            {
                "Features": lower_limit_each_feature.index,
                "lower limit": lower_limit_each_feature.values,
            }
        )
        upper_limit_each_feature_df = pd.DataFrame(
            {
                "Features": upper_limit_each_feature.index,
                "upper limit": upper_limit_each_feature.values,
            }
        )

    if new_feature and return_limits:
        df["num_of_outliers"] = outlier_df.sum(axis=1)
        return (
            df,
            number_of_outlier_each_feature_df,
            lower_limit_each_feature_df,
            upper_limit_each_feature_df,
        )
    elif new_feature:
        # add the new feature indicating this row has outlier data to the data and return.
        df["num_of_outliers"] = outlier_df.sum(axis=1)
        return df, number_of_outlier_each_feature_df

    elif return_limits:
        return lower_limit_each_feature_df, upper_limit_each_feature_df

    return number_of_outlier_each_feature_df


def find_outliers_iqr_method(
    data, new_feature=False, col_name=None, return_limits=False
):
    """Find the outliers in the given dataset.
    :param data: dataset has number of features to find the outliers.
    :param new_feature: If True create a new feature in the dataFrame to indicate this sample has any outlier feature else do nothing.
    :param return_limit: If true return the upper and lower limits.

    :return data with new feature and number of outlier in each features if new_feature is True, else only number of outlier in each features.
    """
    df = data.copy()
    q1_each_features = df.quantile(
        0.25,
    )
    q3_each_features = df.quantile(
        0.75,
    )
    iqr_each_feature = q3_each_features - q1_each_features
    lower_limit_each_feature = q1_each_features - (iqr_each_feature * 1.5)
    upper_limit_each_feature = q3_each_features + (iqr_each_feature * 1.5)

    # find the data is a outlier value or not.
    # print(lower_limit_each_feature, upper_limit_each_feature)
    outlier_df = (df > upper_limit_each_feature) | (df < lower_limit_each_feature)

    # find the number of outliers per feature.
    number_of_outlier_each_feature = outlier_df.sum(axis=0)
    if df.ndim == 1:
        # if the given data is features than handle diffrently. because it has no .values and .index function.
        number_of_outlier_each_feature_df = pd.DataFrame(
            {
                "Feature": [col_name if col_name else "Given feature"],
                "Number of outliers": number_of_outlier_each_feature,
            }
        )
    else:
        number_of_outlier_each_feature_df = pd.DataFrame(
            {
                "Features": number_of_outlier_each_feature.index,
                "Number of outliers": number_of_outlier_each_feature.values,
            }
        )

    # find the upper limit and lower limit Dataframe.
    if df.ndim == 1:
        # if the given data is features than handle diffrently. because it has no .values and .index function.
        lower_limit_each_feature_df = pd.DataFrame(
            {
                "Feature": [col_name if col_name else "Given feature"],
                "lower limit": lower_limit_each_feature,
            }
        )
        upper_limit_each_feature_df = pd.DataFrame(
            {
                "Feature": [col_name if col_name else "Given feature"],
                "upper limit": upper_limit_each_feature,
            }
        )
    else:
        lower_limit_each_feature_df = pd.DataFrame(
            {
                "Features": lower_limit_each_feature.index,
                "lower limit": lower_limit_each_feature.values,
            }
        )
        upper_limit_each_feature_df = pd.DataFrame(
            {
                "Features": upper_limit_each_feature.index,
                "upper limit": upper_limit_each_feature.values,
            }
        )

    if new_feature and return_limits:
        df["num_of_outliers"] = outlier_df.sum(axis=1)
        return (
            df,
            number_of_outlier_each_feature_df,
            lower_limit_each_feature_df,
            upper_limit_each_feature_df,
        )

    elif new_feature:
        # add the new feature indicating this row has outlier data to the data and return.
        df["num_of_outliers"] = outlier_df.sum(axis=1)
        return df, number_of_outlier_each_feature_df

    elif return_limits:
        return lower_limit_each_feature_df, upper_limit_each_feature_df

    return number_of_outlier_each_feature_df


def find_features_with_missing_values(data):
    """
    Find which features having missing values.

    :param data: dataset should be a DataFrame object.
    """
    column_names = data.columns
    feature_with_missing_values = column_names[data.isnull().sum() > 0]
    return feature_with_missing_values


def find_high_correlation(data, threshold):
    """Find the correlation between geatures
    :param data: data set
    :param threshold : threshold to pick the features above than this

    :return dict with correlated features

    # how to read the output
    # eg -- {'feature_7': ['feature_6', 0.9402264038780661], ...
    # feature_7 is correlated with feature_6 with 0.9402264038780661 correlation value
    """
    col_corr_dict = {}
    corr_df = data.corr()
    for row in range(len(corr_df.columns)):
        for col in range(
            row
        ):  # the upper half and lower half is same -- so checking to half way is enough
            if corr_df.iloc[row, col] >= threshold:
                # print(corr_df.columns[col], corr_df.index[row])
                col_corr_dict[corr_df.index[row]] = [
                    corr_df.columns[col],
                    corr_df.iloc[row, col],
                ]  # these two values are highly correlated
    return col_corr_dict



import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from adjustText import adjust_text
from typing import List, Union


# TODO: ADD Docstrings
def biplot_decompose(
    dim1: Union[List[float], np.ndarray],
    dim2: Union[List[float], np.ndarray],
    dim1_weights: Union[List[float], np.ndarray] = None,
    dim2_weights: Union[List[float], np.ndarray] = None,
    labels: List[str] = None,
    hue: List[str] = None,
    palette="plasma",
    ax=None,
    padding: float = 1.2,
    plot_weights=True,
    display: bool = True,
):
    dim1 = np.squeeze(np.array(dim1))
    dim2 = np.squeeze(np.array(dim2))
    dim1_weights = np.squeeze(np.array(dim1_weights))
    dim2_weights = np.squeeze(np.array(dim2_weights))
    to_check = [dim1, dim2]
    if plot_weights:
        to_check.extend([dim1_weights, dim2_weights])
    for input in to_check:
        if len(input.shape) > 1:
            raise ValueError("Input arrays must be 1-dimensional.")
        elif len(input.shape) < 1:
            raise ValueError("input can't be empty")
    if ax is None:
        fig, ax = plt.subplots(figsize=(8, 8))
    else:
        fig = None
    if hue is None:
        sns.scatterplot(x=dim1, y=dim2, ax=ax, color="b")
    else:
        sns.scatterplot(x=dim1, y=dim2, ax=ax, palette=palette)
    ax.set_xlabel("Dimension 1", color="b")
    ax.set_ylabel("Dimension 2", color="b")
    ax.tick_params("y", colors="b")
    ax.tick_params("x", colors="b")

    xlim = max(abs(dim1)) * padding
    ylim = max(abs(dim2)) * padding
    ax.set_xlim(-xlim, xlim)
    ax.set_ylim(-ylim, ylim)
    if not plot_weights:
        sns.despine(left=True, right=True, top=True, bottom=True)
        ax.tick_params(axis="both", which="both", length=0)
        if display:
            plt.show()
        return fig, ax

    ax2 = ax.twinx()
    loading_padding = padding + 0.1
    loading_ylim = np.max(np.abs(dim2_weights)) * loading_padding
    ax2.set_ylim(-loading_ylim, loading_ylim)
    ax2.set_ylabel("Dimension 2 weights", color="r")
    ax2.tick_params("y", colors="r")

    ax3 = ax2.twiny()
    loading_xlim = np.max(np.abs(dim1_weights)) * loading_padding
    ax3.set_xlim(-loading_xlim, loading_xlim)
    ax3.tick_params("x", colors="r")
    ax3.set_xlabel("Dimension 1 weights", color="r")
    texts = []
    for i in range(len(dim1_weights)):
        ax3.arrow(
            0,
            0,
            dim1_weights[i],
            dim2_weights[i],
            head_width=0.01,
            head_length=0.01,
            fc="r",
            ec="r",
            alpha=0.6,
        )
        if labels is not None:
            texts.append(
                ax3.text(
                    dim1_weights[i] * (padding - 0.1),
                    dim2_weights[i] * (padding - 0.1),
                    labels[i],
                    color="r",
                    ha="center",
                    va="center",
                )
            )
    adjust_text(texts)
    sns.despine(left=True, right=True, top=True, bottom=True)

    for cax in [ax, ax2, ax3]:
        cax.tick_params(axis="both", which="both", length=0)
    if display:
        plt.show()
    return fig, (ax, ax2, ax3)


def biplot_paired(
    x_components: np.ndarray,
    y_components: np.ndarray,
    x_weights: np.ndarray,
    y_weights: np.ndarray,
    x_labels: List[str],
    y_labels: List[str],
    n_comps: int,
    hue_x: List[str] = None,
    hue_y: List[str] = None,
    palette: str = "plasma",
    padding: float = 1.2,
    plot_weights: bool = True,
):
    """

    Args:
        x_components:
        y_components:
        x_weights:
        y_weights:
        x_labels:
        y_labels:
        n_comps:
        hue_x:
        hue_y:
        palette:
        padding:
        plot_weights:
        display:

    Returns:

    """
    fig, axes = plt.subplots(
        nrows=n_comps, ncols=n_comps, figsize=(n_comps * 7, n_comps * 7)
    )
    for row_idx in range(n_comps):
        for col_idx in range(n_comps):
            if col_idx < row_idx:  # plot x components
                fig, (ax, ax2, ax3) = biplot_decompose(
                    x_components[:, col_idx],
                    x_components[:, row_idx],
                    x_weights[:, col_idx],
                    x_weights[:, row_idx],
                    x_labels,
                    hue=hue_x,
                    palette=palette,
                    ax=axes[row_idx, col_idx],
                    padding=padding,
                    plot_weights=plot_weights,
                    display=False,
                )
                ax.set_xlabel(f"X Dimension {col_idx+1}", color="b")
                ax.set_ylabel(f"X Dimension {row_idx+1}", color="b")
                ax2.set_ylabel(f"X Dimension {row_idx+1} weights", color="r")
                ax3.set_xlabel(f"X Dimension {col_idx+1} weights", color="r")
            if row_idx == col_idx:  # plot correlations
                cax = axes[row_idx, col_idx]
                if hue_x is None:
                    if hue_y is None:
                        sns.scatterplot(
                            x=x_components[:, row_idx],
                            y=y_components[:, col_idx],
                            ax=cax,
                            color="b",
                        )
                    else:
                        sns.scatterplot(
                            x=x_components[:, row_idx],
                            y=y_components[:, col_idx],
                            ax=cax,
                            hue=hue_y,
                            palette=palette,
                        )
                else:
                    sns.scatterplot(
                        x=x_components[:, row_idx],
                        y=y_components[:, col_idx],
                        ax=cax,
                        hue=hue_x,
                        palette=palette,
                    )

                cax.set_xlabel(f"X Dimension {col_idx+1}")
                cax.set_ylabel(f"Y Dimension {row_idx+1}")
            if col_idx > row_idx:  # plot y components
                fig, (ax, ax2, ax3) = biplot_decompose(
                    y_components[:, col_idx],
                    y_components[:, row_idx],
                    y_weights[:, col_idx],
                    y_weights[:, row_idx],
                    y_labels,
                    hue=hue_y,
                    palette=palette,
                    ax=axes[row_idx, col_idx],
                    padding=padding,
                    plot_weights=plot_weights,
                    display=False,
                )
                ax.set_xlabel(f"Y Dimension {col_idx+1}", color="b")
                ax.set_ylabel(f"Y Dimension {row_idx+1}", color="b")
                ax2.set_ylabel(f"Y Dimension {row_idx+1} weights", color="r")
                ax3.set_xlabel(f"Y Dimension {col_idx+1} weights", color="r")
    plt.tight_layout()
    return fig, axes



import matplotlib.pyplot as plt
from nilearn import datasets, surface, plotting


def plot_surf_stat_4views(
    stat_map,
    vmax: int or float = "auto",
    figsize=(12, 12),
    title: str = None,
):
    fsaverage = datasets.fetch_surf_fsaverage()
    fig, axes = plt.subplots(2, 2, figsize=figsize, subplot_kw={"projection": "3d"})
    mesh_right = surface.load_surf_mesh(fsaverage.pial_right)
    mesh_left = surface.load_surf_mesh(fsaverage.pial_left)
    surf_map_right = surface.vol_to_surf(stat_map, mesh_right)
    surf_map_left = surface.vol_to_surf(stat_map, mesh_left)

    if vmax == "auto":
        vmax = max(surf_map_right.max(), surf_map_left.max())
    elif isinstance(vmax, (float, int)):
        pass
    else:
        raise ValueError("vmax must be a number or str")

    # Plot the right hemisphere (lateral and medial views)
    plotting.plot_surf_stat_map(
        mesh_right,
        surf_map_right,
        hemi="right",
        view="lateral",
        axes=axes[0, 1],
        colorbar=False,
        vmax=vmax,
    )
    plotting.plot_surf_stat_map(
        mesh_right,
        surf_map_right,
        hemi="right",
        view="medial",
        axes=axes[1, 1],
        colorbar=False,
        vmax=vmax,
    )

    # Plot the left hemisphere (lateral and medial views)
    plotting.plot_surf_stat_map(
        mesh_left,
        surf_map_left,
        hemi="left",
        view="lateral",
        axes=axes[0, 0],
        colorbar=False,
        vmax=vmax,
    )
    plotting.plot_surf_stat_map(
        mesh_left,
        surf_map_left,
        hemi="left",
        view="medial",
        axes=axes[1, 0],
        colorbar=False,
        vmax=vmax,
    )

    cax = fig.add_axes(
        [0.92, 0.3, 0.02, 0.5]
    )  # Adjust the position and size of the colorbar
    sm = plt.cm.ScalarMappable(
        cmap=plotting.cm.cold_hot, norm=plt.Normalize(vmin=-vmax, vmax=vmax)
    )
    sm.set_array([])
    fig.colorbar(sm, cax=cax)
    plt.subplots_adjust(left=0.1, right=0.9, bottom=0.1, top=1, wspace=0.0, hspace=-0.4)
    fig.suptitle(title, fontsize=24, y=0.9)

    return fig



import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import StandardScaler, PolynomialFeatures


class NuisanceLinear(BaseEstimator, TransformerMixin):
    def __init__(
        self,
        nuisance_variables: np.ndarray,
        standardize=True,
        line_reg_order=1,
        include_bias=True,
        line_reg_kwargs=None,
    ):
        self.line_reg_kwargs = {} if line_reg_kwargs is None else line_reg_kwargs
        self.line_reg = LinearRegression(**self.line_reg_kwargs)
        self.standardize = standardize
        self.include_bias = include_bias
        self.line_reg_order = line_reg_order

        self.nuisance_variables = nuisance_variables
        if len(self.nuisance_variables.shape) == 1:
            self.nuisance_variables = self.nuisance_variables.reshape(-1, 1)
        if self.standardize:
            scaler = StandardScaler()
            self.nuisance_variables = scaler.fit_transform(self.nuisance_variables)

        self.poly_nuisance_vars = PolynomialFeatures(
            degree=self.line_reg_order, include_bias=self.include_bias
        ).fit_transform(self.nuisance_variables)

    def fit(self, X, y=None):
        self.line_reg.fit(self.poly_nuisance_vars, X)
        return self

    def transform(self, X, y=None):
        predicted_X = self.line_reg.predict(self.poly_nuisance_vars)
        adjusted_X = X - predicted_X
        return adjusted_X

    def fit_transform(self, X, y=None):
        return self.fit(X, y).transform(X, y)



import warnings
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
import numpy as np
import pandas as pd
from sklearn.cross_decomposition import CCA
from sklearn.decomposition import PCA
from sklearn.base import BaseEstimator, TransformerMixin
import seaborn as sns
from matplotlib import pyplot as plt
from scipy.stats import pearsonr, spearmanr, kendalltau
from neuroimager.plotting import density_scatter


warnings.warn("The functions are improved and moved to a separate package: permcca")


class perm_CCA(BaseEstimator, TransformerMixin):
    def __init__(
        self,
        n_comps: int = 1,
        n_perms: int = 1000,
        cca_kwargs: Optional[Dict] = None,
        pca_x: bool = False,
        pca_y: bool = False,
        pca_x_kwargs: Optional[Dict] = None,
        pca_y_kwargs: Optional[Dict] = None,
        random_state: Optional[int] = 42,
    ) -> None:
        """
        n_comp : int, optional, default: 1
            Number of CCA components to use.
        n_perm : int, optional, default: 1000
            Number of permutations to perform.
        random_state : int, RandomState instance or None, optional, default: None
            Seed for random number generator.
        """
        self.n_comps = n_comps
        self.cca_kwargs = cca_kwargs or {}
        self.n_perms = n_perms
        self.random_state = random_state
        self.pca_x = pca_x
        self.pca_y = pca_y
        default_pca_kwargs = {"n_components": 0.95, "whiten": True}
        self.pca_x_kwargs = pca_x_kwargs if pca_x_kwargs else default_pca_kwargs
        self.pca_y_kwargs = pca_y_kwargs if pca_y_kwargs else default_pca_kwargs
        self.cca_model = None
        self.X_c_ = None
        self.Y_c_ = None
        self.orig_rs_ = None
        self.X_loadings_ = None
        self.Y_loadings_ = None

    def _pca_decompose(
        self, raw_x: np.ndarray, raw_y: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        if self.pca_x:
            self.pca_x_model = PCA(**self.pca_x_kwargs)
            self.pca_x_model.fit(raw_x)
            x = self.pca_x_model.transform(raw_x)
        else:
            x = raw_x
        if self.pca_y:
            self.pca_y_model = PCA(**self.pca_y_kwargs)
            self.pca_y_model.fit(raw_y)
            y = self.pca_y_model.transform(raw_y)
        else:
            y = raw_y
        return x, y

    def _cca_decompose(
        self, x: np.ndarray, y: np.ndarray
    ) -> Tuple[CCA, np.ndarray, np.ndarray]:
        x, y = self._pca_decompose(x, y)
        cca = CCA(n_components=self.n_comps, *self.cca_kwargs)
        x_c, y_c = cca.fit_transform(x, y)
        return cca, x_c, y_c

    def _calc_p(self, perm_rs: pd.DataFrame, orig_rs: List[float]) -> List[float]:
        p_vals = []
        for i in range(self.n_comps):
            p_value = ((perm_rs.iloc[:, i]) > abs(orig_rs[i])).sum() / perm_rs.shape[0]
            p_vals.append(p_value)

        return p_vals

    def _permute(self, X: np.ndarray, Y: np.ndarray):
        rng = np.random.default_rng(self.random_state)
        rs = []
        for _ in range(self.n_perms):
            ids = rng.permutation(X.shape[0])
            x_perm = X[ids, :]
            cca_perm, X_perm, Y_perm = self._cca_decompose(x_perm, Y)
            rs.append(
                pd.DataFrame(X_perm).corrwith(
                    pd.DataFrame(Y_perm), method=self.corr_method
                )
            )
        perm_rs = pd.DataFrame(rs)

        return perm_rs

    def permutation_test_cca(
        self, X: np.ndarray, Y: np.ndarray
    ) -> Tuple[CCA, np.ndarray, np.ndarray, List[float], List[float]]:
        """
        Perform a permutation test for Canonical Correlation Analysis (CCA) to assess the significance
        of the correlation between two sets of features.

        Parameters
        ----------
        X : 2D numpy array, shape (n_samples, n_features_X)
            First set of features.
        Y : 2D numpy array, shape (n_samples, n_features_Y)
            Second set of features.
        Returns
        -------
        cca : CCA object
            Fitted CCA model.
        X_c : Transformed X data from CCA.
        Y_c : Transformed Y data from CCA.
        original_correlations : list of float
            Correlation values between the two linear combinations of features.
        p_values : list of float
            List of p-values for each CCA component.
        """
        np.random.seed(self.random_state)

        # Fit the CCA model
        cca, X_c, Y_c = self._cca_decompose(X, Y)

        # Compute the original correlations
        orig_rs = [pearsonr(X_c[:, i], Y_c[:, i])[0] for i in range(self.n_comps)]

        # Perform the permutation test
        perm_rs = self._permute(X, Y)
        p_values = self._calc_p(perm_rs, orig_rs)
        self.cca_model = cca
        self.X_c_ = X_c
        self.Y_c_ = Y_c
        self.orig_rs_ = orig_rs
        return cca, X_c, Y_c, orig_rs, p_values

    def fit(self, X: np.ndarray, Y: np.ndarray):
        """Fits the CCA model on input data."""
        np.random.seed(self.random_state)

        # Fit CCA
        cca, X_c, Y_c = self._cca_decompose(X, Y)
        orig_rs = [pearsonr(X_c[:, i], Y_c[:, i])[0] for i in range(self.n_comps)]

        # Store the fitted CCA and transformed data as class attributes
        self.cca_model = cca
        self.X_c_ = X_c
        self.Y_c_ = Y_c
        self.orig_rs_ = orig_rs
        return self

    def transform(self, X: np.ndarray, Y: np.ndarray):
        """Return the original loadings"""
        if hasattr(self, "pca_x_modarraydel"):
            X_loadings = self.pca_x_model.components_.T @ self.cca_model.x_loadings_
            X_loadings = self.cca_model.x_loadings_
        else:
            X_loadings = self.cca_model.x_loadings_
        if hasattr(self, "pca_y_model"):
            Y_loadings = self.pca_y_model.components_.T @ self.cca_model.y_loadings_
        else:
            Y_loadings = self.cca_model.y_loadings_
        self.x_loadings_ = X_loadings
        self.y_loadings_ = Y_loadings
        return X_loadings, Y_loadings

    def explained_variance_ratio(self) -> Tuple[List[float], List[float]]:
        """
        Calculate the explained variance ratio for both X and Y canonical variates.

        Returns
        -------
        explained_variance_ratio_X: list of float
            Explained variance ratio for each component in the X canonical variates.
        explained_variance_ratio_Y: list of float
            Explained variance ratio for each component in the Y canonical variates.
        """
        if self.X_c_ is None or self.Y_c_ is None:
            raise RuntimeError(
                "Please fit the model before calculating explained variance ratio."
            )

        # Calculate total variance for X and Y canonical variates
        total_variance_X = np.sum(np.var(self.X_c_, axis=0))
        total_variance_Y = np.sum(np.var(self.Y_c_, axis=0))

        # Calculate explained variance ratio for each component
        explained_variance_ratio_X = [
            np.var(self.X_c_[:, i]) / total_variance_X for i in range(self.n_comps)
        ]
        explained_variance_ratio_Y = [
            np.var(self.Y_c_[:, i]) / total_variance_Y for i in range(self.n_comps)
        ]

        return explained_variance_ratio_X, explained_variance_ratio_Y


def plot_cca_scatter(transformed_X, transformed_Y, n_comps=1, **kwargs):
    """
    Plots Canonical Correlation Analysis (CCA) scatter plot.

    Parameters:
    transformed_X (array-like): Transformed X data from CCA.
    transformed_Y (array-like): Transformed Y data from CCA.
    n_comps (int): Number of components to plot.

    Returns:
    None
    """
    fig_size = kwargs.get("fig_size", (10, 10))
    marker = kwargs.get("marker", "o")
    color = kwargs.get("color", "b")
    xlabel = kwargs.get("xlabel", None)
    ylabel = kwargs.get("ylabel", None)
    title_prefix = kwargs.get("title_prefix", "Scatter plot of CCA components")
    layout = kwargs.get("layout", "tight")
    p_values = kwargs.get("p_values", None)
    for i in range(n_comps):
        # plt.scatter(transformed_X[:, i], transformed_Y[:, i], marker=marker, color=color, label=f'Component {i+1}')
        density_scatter(transformed_X[:, i], transformed_Y[:, i], display=False)
        corr_coef = np.corrcoef(transformed_X[:, i], transformed_Y[:, i])[0, 1]
        if kwargs.get("xlabel", False) & kwargs.get("ylabel", False):
            title = f"{title_prefix}: {xlabel} vs {ylabel} pearson r = {corr_coef}"
            if p_values:
                title += f" p = {p_values[i]}"
            plt.title(title)
        else:
            title = f"Component {i + 1}: pearson r = {corr_coef}"
            if p_values:
                title += f" p = {p_values[i]}"
            plt.title(title)
        if not xlabel:
            xlabel = "X Component {}".format(i + 1)
        if not ylabel:
            ylabel = "Y Component {}".format(i + 1)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.tight_layout()
        plt.show()


def plot_cca_weights(
    weights: List or np.ndarray,
    labels: List or np.ndarray,
    ax=False,
    display_thresh=0.2,
    show=False,
    min_font_size=6,
    scaling_factor=50,
    **kwargs,
):
    if isinstance(weights, list):
        weights = np.array(weights)
    elif isinstance(weights, np.ndarray):
        pass
    else:
        raise ValueError("weights must be a list or numpy array")
    if isinstance(labels, list):
        labels = np.array(labels)
    elif isinstance(labels, np.ndarray):
        pass
    else:
        raise ValueError("labels must be a list or numpy array")

    xlim = kwargs.get("xlim", (-1, 1))
    xlabel = kwargs.get("xlabel", "")
    title = kwargs.get("title", "")
    display_title = kwargs.get("display_title", True)
    display_label = kwargs.get("display_label", True)
    sns.set_style("white")
    weights = weights.flatten()
    # get the index of thresholded weights
    idx = np.where(abs(weights) >= display_thresh)[0]
    filtered_weights = weights[idx]
    filtered_labels = labels[idx]
    # Sort weights and labels for X features
    sorted_indices = np.argsort(filtered_weights)
    sorted_weights = filtered_weights[sorted_indices]
    sorted_labels = [filtered_labels[idx] for idx in sorted_indices]
    if not ax:
        fig, ax = plt.subplots(figsize=(8, len(sorted_labels) / 2))
    ax.set_xlim(xlim[0], xlim[1])
    ax.set_ylim(-0.5, len(weights) + 0.5)
    if display_label:
        ax.set_ylabel("Weights")
        ax.set_xlabel(xlabel)
    if display_title:
        ax.set_title(title)
    # perform min-max scalling for weights
    abs_weights = abs(sorted_weights)
    plot_weights = (abs_weights - abs_weights.min()) / (
        abs_weights.max() - abs_weights.min()
    )
    texts = []
    for i, label in enumerate(sorted_labels):
        font_size = max(min_font_size, plot_weights[i] * scaling_factor)
        ax.text(
            0,
            i,
            label + f"  weights {round(sorted_weights[i], 2)}",
            ha="center",
            va="center",
            fontsize=font_size,
        )

    # adjust_text(texts,only_move={ 'text':'X'})
    ax.spines["top"].set_visible(False)
    ax.spines["left"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    ax.set_xticks([])
    ax.set_yticks([])

    if show:
        plt.show()

    return ax


def plot_paired_cca_weights(
    weights_X: np.ndarray or pd.DataFrame,
    weights_Y: np.ndarray or pd.DataFrame,
    x_labels: List = None,
    y_labels: List = None,
    display_thresh: float = 0.2,
    scaling_factor: float = 30,
    subplot_kwargs: dict = None,
):
    """
    Plots Canonical Correlation Analysis (CCA) feature weights.

    Parameters:
    weights_X (array-like): Weights for X features. If passing dataframe, the index represents the feature names.
    weights_Y (array-like): Weights for Y features. If passing dataframe, the index represents the feature names.
    x_labels (list): Labels for X features. Defaults to None. None only works with DataFrame.
    y_labels (list): Labels for Y features. Defaults to None. None only works with DataFrame.
    display_thresh (float): Threshold for displaying feature weights. Defaults to 0.2.
    scaling_factor (float): Scaling factor for font size of displayed weights. Defaults to 30.
    subplot_kwargs (dict, Optional): Default is None, passing to plot_cca_weights() function.
    **kwargs: Additional arguments Controls the figures.
    Returns:
    None
    """
    if isinstance(weights_X, pd.DataFrame):
        x_labels = weights_X.index
        weights_X = weights_X.values
    elif isinstance(weights_X, np.ndarray):
        pass
    else:
        raise TypeError("weights_X must be a numpy array or pandas dataframe.")
    if isinstance(weights_Y, pd.DataFrame):
        y_labels = weights_Y.index
        weights_Y = weights_Y.values
    elif isinstance(weights_Y, np.ndarray):
        pass
    else:
        raise TypeError("weights_Y must be a numpy array or pandas dataframe.")
    if x_labels is not None:
        if not isinstance(x_labels, list):
            try:
                x_labels = list(x_labels)
            except:
                raise TypeError("x_labels must be a list or can be converted to list.")
        if len(x_labels) != weights_X.shape[0]:
            raise ValueError(
                "The length of x_labels must be equal to the number of rows in weights_X"
            )
    else:
        raise ValueError("x_labels must be provided. What are you plotting then?")
    if y_labels is not None:
        if not isinstance(y_labels, list):
            try:
                y_labels = list(y_labels)
            except:
                raise TypeError("y_labels must be a list or can be converted to list.")
        if len(y_labels) != weights_Y.shape[0]:
            raise ValueError(
                "The length of y_labels must be equal to the number of rows in weights_Y"
            )
    else:
        raise ValueError("x_labels must be provided. What are you plotting then?")
    n_comps = weights_X.shape[1]
    for i in range(n_comps):
        fig, (ax1, ax2) = plt.subplots(1, 2)
        plot_cca_weights(
            weights_X[:, i],
            x_labels,
            ax=ax1,
            display_thresh=display_thresh,
            min_font_size=6,
            scaling_factor=scaling_factor,
            xlim=(-1, 1),
            xlabel="X Features",
            title="Component {} X feature weights".format(i + 1),
            **subplot_kwargs,
        )
        plot_cca_weights(
            weights_Y[:, i],
            y_labels,
            ax=ax2,
            display_thresh=display_thresh,
            min_font_size=6,
            scaling_factor=scaling_factor,
            xlabel="Y Features",
            title="Component {} Y feature weights".format(i + 1),
            **subplot_kwargs,
        )
        plt.tight_layout()
        plt.show()
